from chalice import AuthResponse

from chalice_cognito_auth.exceptions import InvalidToken


class UserPoolAuthorizer:
    def __init__(self, decoder, route_selector=None, principal_selector=None):
        self._decoder = decoder

        if route_selector is None:
            route_selector = AllRoutes()
        self._route_selector = route_selector

        if principal_selector is None:
            principal_selector = UsernameSelector()
        self._principal_selector = principal_selector

    def auth_handler(self, auth_request):
        token = auth_request.token
        try:
            claims = self._decoder.decode(token)
            return AuthResponse(
                self._route_selector.get_allowed_routes(claims),
                principal_id=self._principal_selector.get_principal(claims),
            )
        except InvalidToken:
            return AuthResponse(routes=[], principal_id=None)


class RouteSelector:
    def get_allowed_routes(self, claims):
        raise NotImplementedError('get_allowed_routes')


class AllRoutes(RouteSelector):
    def get_allowed_routes(self, claims):
        return ['*']


class PrincipalSelector:
    def get_principal(self, claims):
        raise NotImplementedError('get_principal')


class UsernameSelector:
    def get_principal(self, claims):
        return claims.get('cognito:username')



from collections import namedtuple

from botocore.exceptions import ClientError
from chalice import BadRequestError
from chalice import UnauthorizedError
from chalice import ChaliceViewError


Error = namedtuple('Error', ['cls', 'fmt_str'])
CODE_TO_ERROR = {
    'NotAuthorizedException': Error(UnauthorizedError, '{message}'),
    'UserNotFoundException': Error(UnauthorizedError, '{message}'),
}
DEFAULT_ERROR = Error(ChaliceViewError, '{code}: {message}')


def get_param(body, key, required=False, default=None):
    try:
        return body[key]
    except KeyError as e:
        if required is False:
            return default
        key = e.args[0]
        raise BadRequestError('Missing requred parameter: %s' % key)


def client_error_to_chalice_error(e):
    code = e.response['Error']['Code']
    message = e.response['Error']['Message']
    error = CODE_TO_ERROR.get(code, DEFAULT_ERROR)
    raise error.cls(error.fmt_str.format(code=code, message=message))


def handle_client_errors(fn):
    def wrapped(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except ClientError as e:
            client_error_to_chalice_error(e)
    return wrapped



import time
import json
import urllib.request

from jose import jwt
from jose import jwk
from jose.utils import base64url_decode

from chalice_cognito_auth.exceptions import InvalidToken


class TokenDecoder:
    def __init__(self, key_fetcher, app_client_id, now=None):
        self._key_fetcher = key_fetcher
        self._app_client_id = app_client_id
        if now is None:
            now = time.time
        self._now = now

    def decode(self, token):
        try:
            self._verify(token)
            claims = self._get_claims(token)
            return claims
        except InvalidToken:
            raise
        except:
            raise InvalidToken('Error decoding token')

    def _verify(self, token):
        headers = jwt.get_unverified_headers(token)
        kid = headers['kid']
        key = self._get_key(kid)
        public_key = jwk.construct(key)
        message, encoded_signature = str(token).rsplit('.', 1)
        decoded_signature = base64url_decode(encoded_signature.encode('utf-8'))
        if not public_key.verify(message.encode("utf8"), decoded_signature):
            raise InvalidToken('Signature verification failed')

    def _get_key(self, kid):
        for key in self._key_fetcher.get_keys():
            if key['kid'] == kid:
                return key
        raise InvalidToken('Could not find kid %s' % kid)

    def _get_claims(self, token):
        claims = jwt.get_unverified_claims(token)
        if self._now() > claims['exp']:
            raise InvalidToken('Token expired')
        if claims['aud'] != self._app_client_id:
            raise InvalidToken('Token was not issued for this audience')
        return claims


class KeyFetcher:
    _KEYS_URL = (
        'https://cognito-idp.{region}.amazonaws.com'
        '/{user_pool_id}/.well-known/jwks.json'
    )

    def __init__(self, region, user_pool_id, urlopen=None):
        self._region = region
        self._user_pool_id = user_pool_id
        self._keys = None
        if urlopen is None:
            urlopen = urllib.request.urlopen
        self._urlopen = urlopen

    def get_keys(self):
        if self._keys is None:
            self._keys = self._get_keys()
        return self._keys

    def _get_keys(self):
        url = self._KEYS_URL.format(
            region=self._region,
            user_pool_id=self._user_pool_id,
        )
        return json.loads(self._urlopen(url).read())['keys']



import os
import sys

from chalice import Blueprint
from chalice import Response

from chalice_cognito_auth.exceptions import InvalidAuthHandlerNameError
from chalice_cognito_auth.exceptions import ChallengeError
from chalice_cognito_auth.utils import get_param
from chalice_cognito_auth.utils import handle_client_errors


class BlueprintFactory:
    def _rename_fn(self, name):
        def renamer(fn):
            def wrapped(*args, **kwargs):
                return fn(*args, **kwargs)
            wrapped.__name__ = name
            return wrapped
        return renamer


    def create_blueprint(self, name, authorizer, lifecycle):
        if name in vars(sys.modules[__name__]):
            raise InvalidAuthHandlerNameError(name)

        routes = Blueprint('%s' % __name__)

        @routes.authorizer(name=name)
        @self._rename_fn(name)
        def auth(auth_request):
            return authorizer.auth_handler(auth_request)

        @routes.route('/register', methods=['POST'])
        @handle_client_errors
        def register():
            body = routes.current_request.json_body
            username = get_param(body, 'username', required=True)
            password = get_param(body, 'password', required=True)
            body.pop('username')
            body.pop('password')
            return lifecycle.register(username, password, body)

        @routes.route('/login', methods=['POST'])
        @handle_client_errors
        def login():
            body = routes.current_request.json_body
            username = get_param(body, 'username', required=True)
            password = get_param(body, 'password', required=True)
            try:
                return lifecycle.login(username, password)
            except ChallengeError as e:
                return Response(
                    body=e.params,
                    status_code=401,
                    headers={
                        'Challenge': e.challenge,
                        'Session': e.session,
                    }
                )

        @routes.route('/auth_challenge', methods=['POST'])
        @handle_client_errors
        def auth_challenge():
            body = routes.current_request.json_body
            challenge = get_param(body, 'challenge', required=True)
            session = get_param(body, 'session', required=True)
            params = get_param(body, 'params', required=True)
            return lifecycle.auth_challenge(challenge, session, params)

        @routes.route('/refresh', methods=['POST'])
        @handle_client_errors
        def refresh():
            body = routes.current_request.json_body
            refresh_token = get_param(body, 'refresh_token', required=True)
            return lifecycle.refresh(refresh_token)

        setattr(sys.modules[__name__], name, auth)
        return routes, auth


def _import_chalice_app_if_needed():
    # Chalice isn't loaded in an authorizer because the lambda handler string
    # does not load the app.* file. It loads chalice_cognito_auth.blueprint.*
    # instead. This causes create_blueprint not to get called (since it is
    # called in app.py) and hence it prevents the authorizer handler function
    # from getting injected into this module. This method is called on module
    # load to include app.py (hence calling create_blueprint) and preventing
    # a circular import.
    if 'app' in sys.modules:
        return
    import app


if os.environ.get("AWS_EXECUTION_ENV") is not None:
    _import_chalice_app_if_needed()



import os

import boto3

from chalice_cognito_auth.blueprint import BlueprintFactory
from chalice_cognito_auth.decoder import TokenDecoder
from chalice_cognito_auth.decoder import KeyFetcher
from chalice_cognito_auth.authorizer import UserPoolAuthorizer
from chalice_cognito_auth.exceptions import ChallengeError


class UserPoolHandlerFactory:
    def __init__(self, blueprint_factory=None):
        if blueprint_factory is None:
            blueprint_factory = BlueprintFactory()
        self._blueprint_factory = blueprint_factory

    def create_user_pool_handler(self, app_client_id=None, user_pool_id=None,
                                 region=None, name=None):
        if app_client_id is None:
            app_client_id = os.environ.get('APP_CLIENT_ID')
        if user_pool_id is None:
            user_pool_id = os.environ.get('POOL_ID')
        if region is None:
            region = os.environ.get('AWS_REGION')
        if name is None:
            name = 'UserPoolAuth'
        key_fetcher = KeyFetcher(region, user_pool_id)
        decoder = TokenDecoder(key_fetcher, app_client_id)
        authorizer = UserPoolAuthorizer(decoder)
        cognito = boto3.client('cognito-idp', region_name=region)
        lifecycle = CognitoLifecycle(app_client_id, user_pool_id, cognito)
        blueprint, auth_wrapper = self._blueprint_factory.create_blueprint(
            name, authorizer, lifecycle)
        handler = UserPoolHandler(authorizer, blueprint, auth_wrapper)
        return handler


class UserPoolHandler:
    def __init__(self, authorizer, blueprint, auth_wrapper):
        self._authorizer = authorizer
        self.blueprint = blueprint
        self._auth_wrapper = auth_wrapper

    @property
    def auth(self):
        return self._auth_wrapper

    @property
    def current_user(self):
        request = self.blueprint.current_request
        return request.context.get('authorizer', {}).get('principalId')

    @property
    def pid(self):
        return self.current_user


class CognitoLifecycle:
    def __init__(self, app_client_id, user_pool_id, cognito):
        self._app_client_id = app_client_id
        self._user_pool_id = user_pool_id
        self._cognito = cognito

    def _get_tokens(self, result):
        tokens = {}
        if 'IdToken' in result:
            tokens['id_token'] = result['IdToken']
        if 'RefreshToken' in result:
            tokens['refresh_token'] = result['RefreshToken']
        if 'AccessToken' in result:
            tokens['access_token'] = result['AccessToken']
        if 'TokenType' in result:
            tokens['token_type'] = result['TokenType']
        return tokens

    def _handle_auth_attempt(self, result):
        if 'ChallengeName' in result:
            raise ChallengeError(
	        result['ChallengeName'],
                result['Session'],
                result['ChallengeParameters'],
            )
        result = result['AuthenticationResult']
        if 'AccessToken' in result:
            return self._get_tokens(result)
        return result

    def register(self, username, password, properties):
        user_attributes = [
            {
                'Name': k,
                'Value': v,
            }
            for k, v in properties.items()
        ]
        result = self._cognito.sign_up(
            Username=username,
            Password=password,
            UserAttributes=user_attributes,
            ClientId=self._app_client_id,
        )

    def login(self, username, password):
        result = self._cognito.initiate_auth(
            AuthFlow='USER_PASSWORD_AUTH',
            AuthParameters={
                'USERNAME': username,
                'PASSWORD': password,
            },
            ClientId=self._app_client_id,
        )
        return self._handle_auth_attempt(result)

    def auth_challenge(self, challenge, session, params):
        result = self._cognito.respond_to_auth_challenge(
            ChallengeName=challenge,
            Session=session,
            ChallengeResponses=params,
            ClientId=self._app_client_id,
        )
        return self._handle_auth_attempt(result)

    def refresh(self, refresh_token):
        result = self._cognito.initiate_auth(
            AuthFlow='REFRESH_TOKEN_AUTH',
            AuthParameters={
                'REFRESH_TOKEN': refresh_token,
            },
            ClientId=self._app_client_id,
        )
        result = result['AuthenticationResult']
        if 'AccessToken' in result:
            return self._get_tokens(result)
        return result



import os
import re
import sys
import codecs
from setuptools import setup, find_packages


here = os.path.abspath(os.path.dirname(__file__))
src_dir = os.path.join(here, "src")
sys.path.insert(0, src_dir)


def read(*parts):
    return codecs.open(os.path.join(here, *parts), 'r').read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(
        r"^__version__ = ['\"]([^'\"]*)['\"]",
        version_file, re.M,
    )
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


requires = [
    'python-jose[cryptography]<4.0.0',
    'boto3>=1.9,<2.0',
]


setup_options = dict(
    name='chalice-cognito-auth',
    version=find_version("src", "chalice_cognito_auth", "__init__.py"),
    description='Library for verifying cognito tokens.',
    author='John Carlyle',
    install_requires=requires,
    package_dir={"": "src"},
    packages=find_packages(where="src", exclude=['tests*']),
    license="Apache License 2.0",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 3 - Alpha",
    ],
)


setup(**setup_options)



class InvalidToken(Exception):
    pass


class InvalidAuthHandlerNameError(Exception):
    """InvalidAuthHandlerNameError

    The names of auth handlers must be unique.
    """
    def __init__(self, name):
        self.message = 'Duplicate auth handler name %s found.' % name
        self.duplicate_name = name


class ChallengeError(Exception):
    """ChallengeError

    Raised during login when an additional challenge needs to be passed before
    login.
    """
    def __init__(self, challenge, session, params):
        self.challenge = challenge
        self.session = session
        self.params = params
